

import c4d

from cioc4d.widgets import utils as wutil


class TextAreaGrp(object):
    """A group containing a label, textfield, and optional checkbox.

    TODO: Consolidate with HidableTextField.
    """

    def __init__(self, dialog, **kwargs):
        self.height=kwargs.get("height", 80)
        self.dialog = dialog
        self.text_area_id = None

        self._build()

    def _build(self):

        with wutil.grid_group(self.dialog, cols=1):

            self.text_area_id = self.dialog.register()
            self.dialog.AddMultiLineEditText(
                self.text_area_id,  c4d.BFH_SCALEFIT | c4d.BFV_TOP, initw=0, inith=self.height, style=c4d.DR_MULTILINE_READONLY)

    def set_value(self, value):
        self.dialog.SetString(self.text_area_id, value)

    def get_value(self):
        return self.dialog.GetString(self.text_area_id).strip()




import c4d

from cioc4d.widgets.checkbox_grp import CheckboxGrp
from cioc4d.sections.collapsible_section import CollapsibleSection


class DiagnosticsSection(CollapsibleSection):

    ORDER = 120

    def __init__(self, dialog):
        self.widget = None
        self.fixture_widget = None
        super(DiagnosticsSection, self).__init__(
            dialog, "Diagnostics", collapse=True)

    def build(self):
        self.widget = CheckboxGrp(self.dialog, "Show Tracebacks")
        self.fixture_widget = CheckboxGrp(self.dialog, "Use Fixtures")

    def populate_from_store(self):
        self.widget.set_value(self.dialog.store.show_tracebacks())
        self.fixture_widget.set_value(self.dialog.store.use_fixtures())

    def save_to_store(self):
        store = self.dialog.store
        store.set_show_tracebacks(self.widget.get_value())
        store.set_use_fixtures(self.fixture_widget.get_value())
        store.commit()

    def on_plugin_message(self, widget_id, msg):
        if widget_id in self._store_affectors:
            self.save_to_store()

    def set_store_affectors(self):
        self._store_affectors =  [
            self.widget.checkbox_id,
            self.fixture_widget.checkbox_id,
        ]





import os

import c4d
from ciocore import data as coredata
from cioc4d.widgets.combo_box_grp import ComboBoxGrp
from cioc4d.widgets.text_field_grp import TextFieldGrp
from cioc4d.sections.collapsible_section import CollapsibleSection
 
TAKE_OPTIONS = ["Current", "Marked", "Main"]


class GeneralSection(CollapsibleSection):

    ORDER = 10

    def __init__(self, dialog):
        self.takes_widget = None
        self.title_widget = None
        self.projects_widget = None
        self.destination_widget = None
        self.instance_types_widget = None

        super(GeneralSection, self).__init__(dialog, "General", collapse=False)

    def build(self):

        self.takes_widget = ComboBoxGrp(self.dialog, label="Takes")

        self.title_widget = TextFieldGrp(
            self.dialog, label="Job Title", placeholder="Title in Conductor dashboard")

        self.projects_widget = ComboBoxGrp(
            self.dialog, label="Conductor Project")

        self.destination_widget = TextFieldGrp(
            self.dialog, label="Destination Path", placeholder="Path where renders are saved to")

        self.dialog.AddSeparatorH(inith=0)

        self.instance_types_widget = ComboBoxGrp(
            self.dialog, label="Instance Type", checkbox=True,
            check_label="Preemptible")
        
        self.dialog.Enable(self.takes_widget.combo_box_id, False)

    def populate_from_store(self):

        self.update_combo_items()

        # store is guaranteed to be valid
        store = self.dialog.store
 
        self.instance_types_widget.set_checkbox_value(store.preemptible())
        self.takes_widget.set_items(TAKE_OPTIONS)
        self.takes_widget.set_by_value(store.takes())
        self.title_widget.set_value(store.title())
        self.destination_widget.set_value(store.destination())

    def save_to_store(self):
        store = self.dialog.store
        store.set_preemptible(self.instance_types_widget.get_checkbox_value())
        store.set_takes(self.takes_widget.get_selected_value())
        store.set_title(self.title_widget.get_value())
        store.set_destination(self.destination_widget.get_value())
        store.set_instance_type(
            self.instance_types_widget.get_selected_value())
        store.commit()

    def on_plugin_message(self, widget_id, msg):
        if widget_id == self.dialog.foot_grp.connect_button_id:
            self.update_combo_items()

        # should this be done with a call to super? 
        if widget_id in self._store_affectors:
            self.save_to_store()


    def update_combo_items(self):
        if coredata.valid():
            store = self.dialog.store
            projects = coredata.data()["projects"]
            inst_types = coredata.data()["instance_types"]
            self.projects_widget.set_items(
                projects, default_value=store.project())
            self.instance_types_widget.set_items(
                [val["description"] for val in inst_types], default_value=store.instance_type())
        else:
            self.projects_widget.set_items(["-- Not Connected --"])
            self.instance_types_widget.set_items(["-- Not Connected --"])

    def get_selected_instance_type(self):
        if not coredata.valid():
            return
        inst_types = coredata.data()["instance_types"]
        description = self.instance_types_widget.get_selected_value()
        return next((it for it in inst_types if it["description"] == description), None)

    def resolve(self, expander, **kwargs):
        instance_type = self.get_selected_instance_type()

        result = {
            "job_title": expander.evaluate(self.title_widget.get_value()),
            "output_path": self.destination_widget.get_value(),
            "project": self.projects_widget.get_selected_value(),
            "instance_type": instance_type["name"] if instance_type else "INVALID",
            "preemptible": self.instance_types_widget.get_checkbox_value()
        }
        return result

    def get_preview_affectors(self):
        return [
            self.takes_widget.combo_box_id,
            self.title_widget.text_field_id,
            self.projects_widget.combo_box_id,
            self.destination_widget.text_field_id,
            self.instance_types_widget.combo_box_id,
            self.instance_types_widget.checkbox_id
        ]




# import os
from ciocore import data as coredata

from cioc4d.widgets.combo_box_grp import ComboBoxGrp
from cioc4d.sections.collapsible_section import CollapsibleSection
from ciocore.package_environment import PackageEnvironment


class SoftwareSection(CollapsibleSection):

    ORDER = 20

    def __init__(self, dialog):
        self.host_version_widget = None
        self.renderer_widget = None
        super(SoftwareSection, self).__init__(
            dialog, "Software", collapse=False)

    def build(self):
        self.host_version_widget = ComboBoxGrp(
            self.dialog, label="Cinema 4D Version")
        self.renderer_widget = ComboBoxGrp(self.dialog, label="Renderer")

    def populate_from_store(self):
        self.update_combo_items()

    def save_to_store(self):
        store = self.dialog.store
        store.set_host_version(self.host_version_widget.get_selected_value())
        store.set_renderer_version(self.renderer_widget.get_selected_value())
        store.commit()


    def on_plugin_message(self, widget_id, msg):

        if widget_id == self.dialog.foot_grp.connect_button_id:
            self.update_combo_items()
        elif widget_id == self.host_version_widget.combo_box_id:
            self.update_renderers_combo_items()
        
        if widget_id in self._store_affectors:
            self.save_to_store()

    def update_combo_items(self):
        self.host_version_widget.set_items(["-- Not Connected --"])
        self.renderer_widget.set_items(["-- Not Connected --"])
        if coredata.valid():
            store = self.dialog.store
            software_data = coredata.data()["software"]
            host_names = software_data.supported_host_names()
            if not host_names:
                return
            self.host_version_widget.set_items(
                host_names, default_value=store.host_version())
            self.update_renderers_combo_items()

    def update_renderers_combo_items(self):
        host = self.host_version_widget.get_selected_value()
        software_data = coredata.data()["software"]
        store = self.dialog.store

 
        renderers = software_data.supported_plugins(host)
        if not renderers:
            self.renderer_widget.set_items(["-- Default Renderer --"])
            return
        renderer_names = ["-- Default Renderer --"]
        for renderer in renderers:
            for version in renderer["versions"]:
                renderer_names.append("{} {}".format(
                    renderer["plugin"], version))
        self.renderer_widget.set_items(
            renderer_names, default_value=store.renderer_version())

    def resolve(self, expander, **kwargs):
        extra_env = self.dialog.section(
            "EnvironmentSection").get_entries(expander)

        packages_data = self.get_packages()
        if not packages_data:
            return {"software_package_ids": "INVALID"}

        packages_data["env"].extend(extra_env)
        return {
            "environment": dict(packages_data["env"]),
            "software_package_ids": packages_data["ids"]
        }

    def get_packages(self):
        if not coredata.valid():
            return
        tree_data = coredata.data()["software"]
        paths = []
        host_path = self.host_version_widget.get_selected_value()
        renderer_path = "{}/{}".format(host_path,
                                       self.renderer_widget.get_selected_value())
        paths = [host_path, renderer_path]

        result = {
            "ids": [],
            "env": PackageEnvironment()
        }
        for package in filter(None, [tree_data.find_by_path(path) for path in paths if path]):
            result["ids"].append(package["package_id"])
            result["env"].extend(package)

        return result

    def get_preview_affectors(self):
        return [
            self.host_version_widget.combo_box_id,
            self.renderer_widget.combo_box_id
        ]




import c4d

from cioc4d.widgets.int_field_grp import IntFieldGrp
from cioc4d.sections.collapsible_section import CollapsibleSection


class RetriesSection(CollapsibleSection):

    ORDER = 90

    def __init__(self, dialog):
        self.preempted_widget = None
        self.failed_widget = None

        super(RetriesSection, self).__init__(
            dialog, "Automatic Retries", collapse=True)

    def build(self):
        self.preempted_widget = IntFieldGrp(
            self.dialog, "Retries When Preempted")
        self.failed_widget = IntFieldGrp(self.dialog, "Retries When Failed")

    def populate_from_store(self):
        store = self.dialog.store
        self.preempted_widget.set_value(store.retries_when_preempted())
        self.failed_widget.set_value(store.retries_when_failed())

    def save_to_store(self):
        store = self.dialog.store
        store.set_retries_when_preempted(self.preempted_widget.get_value() )
        store.set_retries_when_failed(self.failed_widget.get_value() )
        store.commit()


    def on_plugin_message(self, widget_id, msg):
        if widget_id in self._store_affectors:
            self.save_to_store()

    def resolve(self, _, **kwargs):
        return {
            "autoretry_policy": {
                "failed": {
                    "max_retries":  self.failed_widget.get_value()
                },
                "preempted": {
                    "max_retries": self.preempted_widget.get_value()
                }
            }
        }

    def get_preview_affectors(self):
        return [
            self.preempted_widget.int_field_id,
            self.failed_widget.int_field_id
        ]




import c4d
import json
from cioc4d.widgets.int_field_grp import IntFieldGrp
from cioc4d.widgets.button_row import ButtonRow
from cioc4d.sections.collapsible_section import CollapsibleSection
from cioc4d.widgets.text_area_grp import TextAreaGrp


class PreviewSection(CollapsibleSection):

    ORDER = 300

    def __init__(self, dialog):
        self.json_widget = None
        self.max_tasks_widget = None
        self.affectors = []

        super(PreviewSection, self).__init__(dialog, "Preview", collapse=True, override_color=c4d.Vector(0.2, 1,0.2))

    def build(self):
        self.max_tasks_widget = IntFieldGrp(self.dialog, "Task Display Limit")
        self.actions_widget = ButtonRow(
            self.dialog, "Show Assets", button_width=120)
            
        self.json_widget = TextAreaGrp(self.dialog, height=200)

    def populate_from_store(self):
        store = self.dialog.store
        self.max_tasks_widget.set_value(store.preview_max_tasks())

    def save_to_store(self):
        store = self.dialog.store
        store.set_preview_max_tasks(
            self.max_tasks_widget.get_value())
        store.commit()

    def on_plugin_message(self, widget_id, msg):
        if widget_id in self.affectors:
            with_assets = widget_id == self.preview_button_id
            self.set_submission_preview(with_assets)

        if widget_id in self._store_affectors:
            self.save_to_store()

    def set_submission_preview(self,  with_assets=False):
        max_tasks = self.max_tasks_widget.get_value()
        submission = self.dialog.calculate_submission(
            max_tasks=max_tasks, with_assets=with_assets)
        self.json_widget.set_value(json.dumps(submission, indent=2))

    def set_affectors(self):
        """
        Get the IDs of widgets that can affect the preview
        """
        self.affectors = []
        self.affectors.append(self.dialog.foot_grp.connect_button_id)

        for section in self.dialog.sections:
            self.affectors += section.get_preview_affectors()

 


    def get_preview_affectors(self):
        """
        Collecting assets is potentially expensive as we need to hit the
        filesystem. For this reason, make the user click a button so they know
        what to expect.
        """
        return [
            self.max_tasks_widget.int_field_id,
            self.actions_widget.button_ids[0]
        ]

    @property
    def preview_button_id(self):
        return self.actions_widget.button_ids[0]





import c4d
from cioc4d.sections.collapsible_section import CollapsibleSection
from cioc4d.widgets.key_value_widget import KeyValueHeader, KeyValueWidget


class MetadataSection(CollapsibleSection):
    ORDER=70
    def __init__(self, dialog):
        self.list_grp_id = None
        self.widgets = []
        super(MetadataSection, self).__init__(
            dialog, "Metadata", collapse=True)

    def build(self):
        self.header_row = KeyValueHeader(self.dialog)

        self.list_grp_id = self.dialog.register()

        self.dialog.GroupBegin(
            id=self.list_grp_id,
            flags=c4d.BFH_SCALEFIT,
            title="",
            cols=1,
            groupflags=0)

        self.dialog.GroupEnd()  # main

    def populate_from_store(self):
        self._update_widgets(self.dialog.store.metadata())

    def save_to_store(self):
        store = self.dialog.store
        value_tuples = [widget.get_values() for widget in self.widgets ] 
        store.set_metadata(  [t for t in value_tuples if t[0].strip()]  )
        store.commit()

    def on_plugin_message(self, widget_id, msg):
        if widget_id == self.header_row.add_button_id:  # clear list
            self.add_entry()
        else :
            self.maybe_remove_entry(widget_id)
 
        if widget_id in self._store_affectors:
            self.save_to_store()

    def add_entry(self):
        values = [widget.get_values() for widget in self.widgets]
        values.append(("", ""))
        self._update_widgets(values)

    def maybe_remove_entry(self, widget_id):
        keep_widgets = [
            w for w in self.widgets if w.delete_btn_id != widget_id]
        if len(keep_widgets) < len(self.widgets):

            values = [widget.get_values() for widget in keep_widgets]
            self._update_widgets(values)
            return True
        return False

    def _update_widgets(self, values):
        self.widgets = []
        self.dialog.LayoutFlushGroup(self.list_grp_id)
        for value_tuple in values:
            self.widgets.append(
                KeyValueWidget(self.dialog, *value_tuple))
        self.dialog.LayoutChanged(self.list_grp_id)


    def resolve(self, expander, **kwargs):
        metadata = {}
        for widget in self.widgets:
            name, value = widget.get_values()
            metadata[name]=value
        return {"metadata": expander.evaluate(metadata) }
 
    def get_preview_affectors(self):
        result = []
        result.append(self.header_row.add_button_id)
        for widget in self.widgets:
            result += [widget.name_field_id, widget.value_field_id, widget.delete_btn_id]
        return result




import c4d
# import cioc4d.const as k
from cioc4d.widgets import utils as wutil


class HidableTextFieldGrp(object):
    """A widget containing a label, a textfield, and checkbox to hide it.
    
    Also has an optional extra checkbox.
    """

    def __init__(self, dialog, label, **kwargs):
        self.dialog = dialog
        self.label = label
        self.helptext = kwargs.get("placeholder")
        self.group_id = None
        self.label_id = None
        self.checkbox_id = None
        self.text_field_id = None

        self.extra_checkbox_label = kwargs.get("extra_checkbox_label", None)
        self.extra_checkbox_id = None
        self.has_extra_checkbox = self.extra_checkbox_label != None
        self._build()

    def _build(self):

        cols = 4 if self.has_extra_checkbox else 3
        with wutil.grid_group(self.dialog, cols=cols) as group_id:
            self.group_id = group_id

            self.label_id = wutil.dotted_label(self.dialog, self.label)

            self.checkbox_id = self.dialog.register()
            self.dialog.AddCheckbox(
                id=self.checkbox_id, flags=c4d.BFH_LEFT, initw=30,  inith=0,   name="")

            self.text_field_id = self.dialog.register()
            self.dialog.AddEditText(
                self.text_field_id, c4d.BFH_SCALEFIT | c4d.BFV_TOP,
                editflags=c4d.EDITTEXT_HELPTEXT)

            if self.helptext:
                self.dialog.SetString(
                    self.text_field_id, self.helptext, flags=c4d.EDITTEXT_HELPTEXT)

            if self.has_extra_checkbox:
                self.extra_checkbox_id = self.dialog.register()
                self.dialog.AddCheckbox(
                    id=self.extra_checkbox_id, flags=c4d.BFH_RIGHT, initw=0,  inith=0, name=self.extra_checkbox_label)

        self.set_visible(True)

    def set_value(self, value):
        self.dialog.SetString(self.text_field_id, value)

    def get_value(self):
        return self.dialog.GetString(self.text_field_id).strip()

    def set_visible(self, value=None):
        if value != None:
            self.dialog.SetBool(self.checkbox_id, value)
        vis = self.dialog.GetBool(self.checkbox_id)
        self.dialog.HideElement(self.text_field_id, not vis)
        if self.has_extra_checkbox:
            self.dialog.HideElement(self.extra_checkbox_id, not vis)
        self.dialog.LayoutChanged(self.group_id)

    def get_visible(self):
        return self.dialog.GetBool(self.checkbox_id)

    def get_extra_check_value(self):
        if self.has_extra_checkbox:
            return self.dialog.GetBool(self.extra_checkbox_id)

    def set_extra_check_value(self, value):
        if self.has_extra_checkbox:
            return self.dialog.SetBool(self.extra_checkbox_id, value)

    def set_extra_check_enabled(self, enable):
        if self.has_extra_checkbox:
            self.dialog.Enable(self.extra_checkbox_id, enable)
            self.dialog.LayoutChanged(self.group_id)





import c4d
from cioc4d.widgets import utils as wutil


class ButtonRow(object):
    """A group containing any number of buttons in a row."""

    def __init__(self, dialog, *labels, **kwargs):
        self.label =  kwargs.get("label", "")
        self.dialog = dialog
        self.labels = labels
        self.button_ids = []
        self.num_buttons = len(labels)
        self._build(**kwargs)

    def _build(self, **kwargs):
        button_width = kwargs.get("button_width")
        with wutil.grid_group(self.dialog, cols=2):
            wutil.dotted_label(self.dialog, self.label)
            with wutil.grid_group(self.dialog, cols=self.num_buttons):
                for label in self.labels:
                    this_id = self.dialog.register()
                    self.button_ids.append(this_id)
                    if button_width:
                        self.dialog.AddButton(
                            this_id, c4d.BFH_FIT, initw=button_width,  name=label)
                    else:
                        self.dialog.AddButton(
                            this_id, c4d.BFH_SCALEFIT,  name=label)





import c4d
from cioc4d.widgets import utils as wutil


class CheckboxGrp(object):
    """A group containing a label and checkbox"""

    def __init__(self, dialog, label ):
        self.label = label
        self.dialog = dialog
        self.checkbox_id = None
        self._build()
 
    def _build(self):
        with wutil.grid_group(self.dialog, cols=2):
            wutil.dotted_label(self.dialog, self.label)
            self.checkbox_id = self.dialog.register()
            self.dialog.AddCheckbox(
                    self.checkbox_id, c4d.BFH_SCALEFIT , initw=0, inith=0 , name="")

    def get_value(self):
        return self.dialog.GetBool(self.checkbox_id) 

    def set_value(self, value):
        self.dialog.SetBool(self.checkbox_id, value) 






import c4d
from cioc4d.widgets import utils as wutil


class TextGrp(object):
    """A group containing a label, and static text.
    
    TODO: Label should not be kwargs.
    """

    def __init__(self, dialog, **kwargs):
        self.dialog = dialog
        self.label = kwargs.get("label", "")
        self.text_id = None
        self._build()

    def _build(self):
        with wutil.grid_group(self.dialog, cols=2):
            wutil.dotted_label(self.dialog, self.label)

            self.text_id = self.dialog.register()
            self.dialog.AddStaticText(
                self.text_id,
                c4d.BFH_SCALEFIT | c4d.BFV_TOP)

    def set_value(self, value):
        self.dialog.SetString(self.text_id, value)

    def get_value(self):
        return self.dialog.GetString(self.text_id)




import c4d
import re
from cioc4d.sections.collapsible_section import CollapsibleSection
from cioc4d.widgets.hidable_text_field_grp import HidableTextFieldGrp


class NotificationSection(CollapsibleSection):
    ORDER = 75

    def __init__(self, dialog):
        self.widget = None

        super(NotificationSection, self).__init__(
            dialog, "Notifications", collapse=True)

    def build(self):

        self.widget = HidableTextFieldGrp(
            self.dialog, "Email Addresses")

    def populate_from_store(self):
        store = self.dialog.store
        self.widget.set_value(store.emails())
        self.widget.set_visible(store.use_emails())

    def save_to_store(self):
        store = self.dialog.store
        store.set_emails(self.widget.get_value())
        store.set_use_emails(self.widget.get_visible())
        store.commit()

    def on_plugin_message(self, widget_id, msg):
        if widget_id == self.widget.checkbox_id:
            self.widget.set_visible()
            
        if widget_id in self._store_affectors:
            self.save_to_store()

    def resolve(self, _, **kwargs):
        if not self.widget.get_visible():
            return {}
        return {"notify": [a for a in re.split(r"[, ]+",  self.widget.get_value()) if a]}

    def get_preview_affectors(self):
        return [self.widget.text_field_id, self.widget.checkbox_id]





import c4d
import cioc4d.const as k
from contextlib import contextmanager

def dotted_label(dialog, label):
    """Label, correct width, with dotted filler."""
    elem_id = dialog.register()
    dialog.AddStaticText(
        id=elem_id,
        initw=k.LABEL_WIDTH, inith=k.LABEL_HEIGHT,
        name=label,
        borderstyle=c4d.BORDER_TEXT_DOTTED,
        flags=c4d.BFH_FIT | c4d.BFV_TOP)
    return elem_id



@contextmanager
def grid_group(dialog, **kwargs):
    """Group to contain gadgets for one widget.
    """
    grp_id = dialog.register()
    grp = dialog.GroupBegin(
        id=grp_id,
        flags=c4d.BFH_SCALEFIT,
        cols=kwargs.get("cols", 2),
        groupflags=0)
    if grp:
        yield grp_id
    dialog.GroupEnd()





from cioc4d.widgets.hidable_text_field_grp import HidableTextFieldGrp
from cioc4d.sections.collapsible_section import CollapsibleSection


class AutosaveSection(CollapsibleSection):

    ORDER = 100

    def __init__(self, dialog):
        self.widget = None
        super(AutosaveSection, self).__init__(
            dialog, "Autosave", collapse=True)

    def build(self):
        self.widget = HidableTextFieldGrp(
            self.dialog, "Autosave Filename", extra_checkbox_label="Cleanup")

    def populate_from_store(self):
        store = self.dialog.store
        self.widget.set_visible(store.use_autosave())
        self.widget.set_value(store.autosave_filename())
        self.widget.set_extra_check_value(store.autosave_cleanup())
        self._disable_on_use_daemon()

    def save_to_store(self):
        store = self.dialog.store
        store.set_use_autosave(self.widget.get_visible())
        store.set_autosave_filename(self.widget.get_value())
        store.set_autosave_cleanup(self.widget.get_extra_check_value())
        store.commit()

    def on_plugin_message(self, widget_id, msg):
        if widget_id == self.widget.checkbox_id:
            self.widget.set_visible()
        elif widget_id == self.dialog.section("UploadOptionsSection").use_daemon_widget.checkbox_id:
            self._disable_on_use_daemon()

        if widget_id in self._store_affectors:
            self.save_to_store()

    def _disable_on_use_daemon(self):
        use_daemon = self.dialog.section(
            "UploadOptionsSection").use_daemon_widget.get_value()
        self.widget.set_extra_check_enabled(not use_daemon)

    def set_store_affectors(self):
        self._store_affectors = [
            self.widget.checkbox_id,
            self.widget.text_field_id,
            self.widget.extra_checkbox_id
        ]




"""
Encapsulation of a c4d container.

The purpose of the store is to make it possible to persist the state of the
dialog (settings) in the scene file. 

It is NOT the single source of truth while dialog is open. The widgets are.

So the flow is:

Starting from a new scene or loaded scene or maybe dialog is open already as
part of saved layout.
- User opens dialog:
- Check if scene has an instance of the conductor_container.
- If not, make one and reset it to factory defaults.
- If it does have one, populate the UI from it.

If dialog open and user loads a scene, then a EV MSG CHANGE event will fire, so
we want to repopulate the UI with the new container (if the loaded scene has
one).

Problem is, the same event fires in many situations. So in order to determine
whether the scene really did change, we maintain a MODIFIED timestamp on the
store object, AND in the container. MODIFIED is updated w Each time event fires we compare those
timestamps. In theory, the only time they will be different is when a new scene
is loaded.

"""
import c4d
import os
import json
import cioc4d.const as k
import time
from ciocore.gpath import Path
DEFAULT_TASK_TEMPLATE = 'Commandline -render "<docfile>" -frame <start> <end> -oimage "<docdir>/renders/test."'
DEFAULT_AUTOSAVE_TEMPLATE = "<docdir>/cio_<docname>"
DEFAULT_TITLE = "C4D <docnamex> <takename>"
# IDS: NOTE Always add to the end - Don't insert! Don't ever reorder or remove
# entries, even if an attribute becomes redundant. If you do it will make old
# scenes incompatible.
X = 2000
TAKES = X = X+1
TITLE = X = X+1
PROJECT = X = X+1
DESTINATION = X = X+1
EXTRA_ASSETS = X = X+1
INSTANCE_TYPE = X = X+1
PREEMPTIBLE = X = X+1
CHUNK_SIZE = X = X+1
USE_CUSTOM_RANGE = X = X+1
CUSTOM_RANGE = X = X+1
USE_SCOUT_FRAMES = X = X+1
SCOUT_FRAMES = X = X+1
TASK_TEMPLATE = X = X+1
EXTRA_ENVIRONMENT = X = X+1
METADATA = X = X+1
USE_UPLOAD_DAEMON = X = X+1
UPLOAD_ONLY = X = X+1
RETRIES_WHEN_PREEMPTED = X = X+1
RETRIES_WHEN_FAILED = X = X+1
USE_AUTOSAVE = X = X+1
AUTOSAVE_FILENAME = X = X+1
AUTOSAVE_CLEANUP = X = X+1
LOCATION_TAG = X = X+1
SHOW_TRACEBACKS = X = X+1
HOST_VERSION = X = X+1
RENDERER_VERSION = X = X+1
EMAILS = X = X+1
USE_EMAILS = X = X+1
PREVIEW_MAX_TASKS = X = X+1
USE_FIXTURES = X = X+1
MODIFIED = X = X+1


class ConductorStore(object):
    """
    The store is used to persist a submission recipe in the scene file, and to
    repopulate the dialog when it's rebuilt.
    """
    CONTAINER_ID = k.PLUGIN_ID

    def __init__(self):

        self._modified = None
        self._container = None
        doc = c4d.documents.GetActiveDocument()
        if doc is None:  # Not even sure if its possible for doc to be None
            return
        doc_container = doc.GetDataInstance()
        self._container = doc_container[self.CONTAINER_ID]

        if self._container is None:
            self._container = c4d.BaseContainer()
            self.reset()
            self.commit()

    def on_scene_change(self):
        """
        If there's a scene chamnged event, it may be because the user loaded a
        new scene. In that case we have to use that new scene's
        conductor_container - if it has one.
        """

        # get the potentially different document
        doc = c4d.documents.GetActiveDocument()
        if doc is None:  # Not even sure if its possible for doc to be None
            return
        doc_container = doc.GetDataInstance()
        conductor_container = doc_container[self.CONTAINER_ID]
        if conductor_container is None:

            # Since this new doc has no conductor_container of its own, we use
            # the store's container. We reset its values and then commit it
            # (attach it to the new doc).
            self.reset()
            self.commit()
            return True

        else:
            # There is a conductor_container. If it's timestamp is different to
            #  the store, then the event must have been triggered from loading a new file.
            if conductor_container[MODIFIED] != self._modified:

                #  We set this container as the store's member data and sync
                #  it's timestamp.
                self._container = conductor_container
                self._modified = conductor_container[MODIFIED]
                return True
            else:
                return False

    def commit(self):
        timestamp = int(time.time() * 1000)
        doc = c4d.documents.GetActiveDocument()
        self.set_modified(timestamp)
        self._modified = timestamp

        doc[self.CONTAINER_ID] = self._container

    def clear(self):
        self._container.FlushAll()

    # accessors`
    def modified(self):
        return self._container[MODIFIED]

    def set_modified(self, value):
        self._container.SetInt64(MODIFIED, value)

    def takes(self):
        return self._container[TAKES]

    def set_takes(self, value):
        self._container.SetString(TAKES, value)

    def title(self):
        return self._container[TITLE]

    def set_title(self, value):
        self._container.SetString(TITLE, value)

    # project

    def project(self):
        return self._container[PROJECT]

    def set_project(self, value):
        self._container.SetString(PROJECT, value)

    # We will store the description only
    def instance_type(self):
        return self._container[INSTANCE_TYPE]

    def set_instance_type(self, value):
        self._container.SetString(INSTANCE_TYPE, value)

    def preemptible(self):
        return self._container[PREEMPTIBLE]

    def set_preemptible(self, value):
        self._container.SetBool(PREEMPTIBLE, value)

    def destination(self):
        return self._container[DESTINATION]

    def set_destination(self, value):
        self._container.SetString(DESTINATION, value)

    def chunk_size(self):
        return self._container[CHUNK_SIZE]

    def set_chunk_size(self, value):
        self._container.SetInt32(CHUNK_SIZE, value)

    def use_custom_range(self):
        return self._container[USE_CUSTOM_RANGE]

    def set_use_custom_range(self, value):
        self._container.SetBool(USE_CUSTOM_RANGE, value)

    def custom_range(self):
        return self._container[CUSTOM_RANGE]

    def set_custom_range(self, value):
        self._container.SetString(CUSTOM_RANGE, value)

    def use_scout_frames(self):
        return self._container[USE_SCOUT_FRAMES]

    def set_use_scout_frames(self, value):
        self._container.SetBool(USE_SCOUT_FRAMES, value)

    def scout_frames(self):
        return self._container[SCOUT_FRAMES]

    def set_scout_frames(self, value):
        self._container.SetString(SCOUT_FRAMES, value)

    def task_template(self):
        return self._container[TASK_TEMPLATE]

    def set_task_template(self, value):
        self._container.SetString(TASK_TEMPLATE, value)

    def extra_environment(self):
        return json.loads(self._container[EXTRA_ENVIRONMENT]) or []

    def set_extra_environment(self, obj={}):
        self._container.SetString(EXTRA_ENVIRONMENT,  json.dumps(obj))

    def metadata(self):
        return json.loads(self._container[METADATA]) or []

    def set_metadata(self, obj):
        self._container.SetString(METADATA,  json.dumps(obj))

    def use_upload_daemon(self):
        return self._container[USE_UPLOAD_DAEMON]

    def set_use_upload_daemon(self, value):
        self._container.SetBool(USE_UPLOAD_DAEMON, value)

    def upload_only(self):
        return self._container[UPLOAD_ONLY]

    def set_upload_only(self, value):
        self._container.SetBool(UPLOAD_ONLY, value)

    def retries_when_preempted(self):
        return self._container[RETRIES_WHEN_PREEMPTED]

    def set_retries_when_preempted(self, value):
        self._container.SetInt32(RETRIES_WHEN_PREEMPTED, value)

    def retries_when_failed(self):
        return self._container[RETRIES_WHEN_FAILED]

    def set_retries_when_failed(self, value):
        self._container.SetInt32(RETRIES_WHEN_FAILED, value)

    def use_autosave(self):
        return self._container[USE_AUTOSAVE]

    def set_use_autosave(self, value):
        self._container.SetBool(USE_AUTOSAVE, value)

    def autosave_filename(self):
        return self._container[AUTOSAVE_FILENAME]

    def set_autosave_filename(self, value):
        self._container.SetString(AUTOSAVE_FILENAME, value)

    def autosave_cleanup(self):
        return self._container[AUTOSAVE_CLEANUP]

    def set_autosave_cleanup(self, value):
        self._container.SetBool(AUTOSAVE_CLEANUP, value)

    def location_tag(self):
        return self._container[LOCATION_TAG]

    def set_location_tag(self, value):
        self._container.SetString(LOCATION_TAG, value)

    def show_tracebacks(self):
        return self._container[SHOW_TRACEBACKS]

    def set_show_tracebacks(self, value):
        self._container.SetBool(SHOW_TRACEBACKS, value)

    def use_fixtures(self):
        return self._container[USE_FIXTURES]

    def set_use_fixtures(self, value):
        self._container.SetBool(USE_FIXTURES, value)

    def host_version(self):
        return self._container[HOST_VERSION]

    def set_host_version(self, value):
        self._container.SetString(HOST_VERSION, value)

    def use_emails(self):
        return self._container[USE_EMAILS]

    def set_use_emails(self, value):
        self._container.SetBool(USE_EMAILS, value)

    def emails(self):
        return self._container[EMAILS]

    def set_emails(self, value):
        self._container.SetString(EMAILS, value)

    def renderer_version(self):
        return self._container[RENDERER_VERSION]

    def set_renderer_version(self, value):
        self._container.SetString(RENDERER_VERSION, value)

    def preview_max_tasks(self):
        return self._container[PREVIEW_MAX_TASKS]

    def set_preview_max_tasks(self, value):
        self._container.SetInt32(PREVIEW_MAX_TASKS, value)

    def assets(self):
        return [item[1] for item in self._container.GetContainerInstance(EXTRA_ASSETS)]

    def set_assets(self, assets=[]):

        assets_container = self._container.GetContainerInstance(EXTRA_ASSETS)
        assets_container.FlushAll()
        for i, asset in enumerate(assets):
            assets_container.SetFilename(i, asset)

    def reset(self):

        self.clear()

        assets_container = c4d.BaseContainer()
        self._container.SetContainer(EXTRA_ASSETS, assets_container)

        self.set_takes("Main")

        self.set_title(DEFAULT_TITLE)

        self.set_project("default")
        self.set_instance_type("unknowable")
        self.set_preemptible(True)
        self.set_chunk_size(1)
        self.set_use_custom_range(False)
        self.set_custom_range("1-10")
        self.set_use_scout_frames(True)
        self.set_scout_frames("auto:3")

        self.set_destination(os.path.join(_get_destination_dir()))

        self.set_task_template(DEFAULT_TASK_TEMPLATE)
        self.set_extra_environment()
        self.set_metadata([
            ("submitter", "cioc4d {}".format(k.VERSION))
        ])

        self.set_use_upload_daemon(False)
        self.set_upload_only(False)
        self.set_retries_when_preempted(1)
        self.set_retries_when_failed(1)
        self.set_use_autosave(True)
        self.set_autosave_filename(DEFAULT_AUTOSAVE_TEMPLATE)
        self.set_autosave_cleanup(True)

        self.set_preview_max_tasks(5)
        self.set_location_tag("")
        self.set_emails("artist@example.com, producer@example.com")
        self.set_use_emails(False)

        self.set_show_tracebacks(False)
        self.set_use_fixtures(False)

        self.set_assets()

        self.set_host_version("unknown")
        self.set_renderer_version("default")


def _get_destination_dir():
    """Provide destination path - called by reset().

    When resetting, it means either: 
    1. A new doc was created.
    2. A doc was opened (and had no conductor container).
    3. A doc existed when the dialig was created (and had no conductor container). 

    In cases 2 and 3, the image path could be valid in render settings, so we try to use it.
    In case 1, we generate a destination path from the <docpath>/renders.
    """
    doc = c4d.documents.GetActiveDocument()

    if doc is None:
        return
    render_data = doc.GetActiveRenderData()
    render_file = render_data[c4d.RDATA_PATH]
    doc_path = doc.GetDocumentPath()

    try:
        render_path = os.path.dirname(
            Path(render_file).posix_path()) if render_file else None
    except BaseException:
        render_path = None

    result = render_path if render_path else os.path.join(
        doc_path, "renders") if doc_path else os.path.expanduser("~/renders")

    return result



import abc
import importlib
import inspect
import os
import re
import c4d
from ciocore.expander import Expander
from cioc4d import asset_cache
from ciocore.gpath import Path
from ciocore.gpath_list import GLOBBABLE_REGEX, PathList
from ciocore.validator import Validator, ValidationError


class ValidateUploadDaemon(Validator):
    def run(self, _):
        dialog = self._submitter
        use_daemon = dialog.section(
            "UploadOptionsSection").use_daemon_widget.get_value()
        if not use_daemon:
            return

        msg = "This submission expects an uploader daemon to be running.\n After you press submit you can open a shell and type: 'conductor uploader'"

        location = (dialog.section(
            "LocationSection").widget.get_value() or "").strip()
        if location:
            msg = "This submission expects an uploader daemon to be running and set to a specific location tag.\nAfter you press OK you can open a shell and type: 'conductor uploader --location {}'".format(
                location)

        self.add_notice(msg)
        # By also writing the message to the console, the user can copy paste
        # `conductor uploader --location blah`.
        c4d.WriteConsole(msg)


class ValidateTaskCount(Validator):
    def run(self, _):
        dialog = self._submitter
        count = dialog.section("InfoSection").frame_count
        if count > 1000:
            self.add_notice(
                "This submission contains over 1000 tasks ({}). Are you sure this is correct?".format(count))


class ValidateDestinationDirectoryClash(Validator):
    def run(self, _):
        dialog = self._submitter
        context = dialog.get_context()
        expander = Expander(safe=True, **context)
        dest = expander.evaluate(dialog.section(
            "GeneralSection").destination_widget.get_value())

        dest_path = Path(dest).posix_path(with_drive=False)
        # print "dest_path:" dest_path

        for gpath in asset_cache.data(self._submitter):
            asset_path = gpath.posix_path(with_drive=False)
            if asset_path.startswith(dest_path):
                c4d.WriteConsole("Some of your upload assets exist in the specified output destination directory\n. {} contains {}".format(
                    dest_path, asset_path))
                self.add_error(
                    "The destination directory for rendered output contains assets that are in the upload list. This can cause your render to fail. See the script editor for details.")
                break
            if dest_path.startswith(asset_path):
                c4d.WriteConsole("You are trying to upload a directory that contains your destination directory.\n. {} contains {}".format(
                    asset_path, dest_path))
                self.add_error(
                    "One of your assets is a directory that contains the specified output destination directory. This will cause your render to fail. See the script editor for details.")
                break


class ValidateRenderToDestinationDirectory(Validator):
    def run(self, _):
        dialog = self._submitter
        context = dialog.get_context()
        expander = Expander(safe=True, **context)
        dest = expander.evaluate(dialog.section(
            "GeneralSection").destination_widget.get_value())
        dest_path = Path(dest).posix_path(with_drive=False)

        task_template = dialog.section("TaskSection").widget.get_value()
        overrides_image = "-oimage" in task_template
        overrides_multipass = "-omultipass" in task_template

        document = c4d.documents.GetActiveDocument()
        render_data = document.GetActiveRenderData()

        save_enabled = render_data[c4d.RDATA_GLOBALSAVE]

        if not save_enabled:
            self.add_warning(
                "Saving images is not enabled in Render settings. Your render will most likely fail.")

        do_image_save = render_data[c4d.RDATA_SAVEIMAGE]
        do_multipass_save = render_data[c4d.RDATA_MULTIPASS_SAVEIMAGE] and render_data[c4d.RDATA_MULTIPASS_ENABLE]

        if overrides_image:
            self.add_notice(
                "Your task template overrides the Render Settings image path with the '-oimage' flag. This is fine, but please ensure it points to a writable directory on Conductor's render node. It should be somewhere below the Destination Directory as specified in the general section of the Conductor dialog.")
        elif do_image_save:
            try:
                image_path = Path(render_data[c4d.RDATA_PATH]).posix_path(
                    with_drive=False)

                if not image_path.startswith(dest_path):
                    self.add_warning(
                        "The image path you've set in Render Settings does not point to a writable location. If you continue, your submission will probably fail. Images should be written somewhere below the Destination Directory as specified in the general section of the Conductor dialog. You should change either the path in Render Settings, or the Conductor destination directory.")
            except ValueError:
                self.add_warning(
                    "The image path you've set in Render Settings is not a valid path. If you continue, your submission will probably fail. Images should be written somewhere below the Destination Directory as specified in the general section of the Conductor dialog. You should change either the path in Render Settings, or the Conductor destination directory.")
        else:
            self.add_notice(
                "This file does not specify a single-pass output image.")

        if overrides_multipass:
            self.add_notice(
                "Your task template overrides the Render Settings multipass image path with the '-omultipass' flag. This is fine, but please ensure it points to a writable directory on Conductor's render node. It should be somewhere below the Destination Directory as specified in the general section of the Conductor dialog.")
        elif do_multipass_save:
            try:
                multipass_path = Path(
                    render_data[c4d.RDATA_MULTIPASS_FILENAME]).posix_path(with_drive=False)

                if not multipass_path.startswith(dest_path):
                    self.add_warning("The multipass image path you've set in Render Settings does not point to a writable location. If you continue, you will not be able to download your multipass images. Images should be written somewhere below the Destination Directory as specified in the general section of the Conductor dialog. You should change either the multipass path in Render Settings, or the Conductor destination directory.")
            except ValueError:
                self.add_warning(
                    "The multipass path you've set in Render Settings is not a valid path. If you continue, your submission will probably fail. Images should be written somewhere below the Destination Directory as specified in the general section of the Conductor dialog. You should change either the path in Render Settings, or the Conductor destination directory.")
        else:
            pass
            # The notice below seems pointless because multipass is (probably) not a normal thing in c4d.
            # self.add_notice("This file does not specify a multipass output image.")


class ValidateMissingExtraAssets(Validator):

    def run(self, _):

        missing = []
        for gpath in self._submitter.section("AssetsSection").pathlist:
            pp = gpath.posix_path()
            if not os.path.exists(pp):
                missing.append(pp)

        if missing:
            self.add_warning(
                "Some of the assets specified in the Extra Assets section do not exist on disk. See the console for details. You can continue if you don't need them.")

            c4d.WriteConsole("----- Conductor Asset Validation -------\n")
            for asset in missing:
                c4d.WriteConsole("Missing: {}\n".format(asset))


# Implement more validators here
####################################
####################################


def run(dialog, submitting=True):

    errors, warnings, notices = _run_validators(dialog)

    if errors:
        msg = ""
        msg += "Some errors would cause the submission to fail:\n\n" + \
            "\n".join(errors)+"\n"
        c4d.gui.MessageDialog(msg,  type=c4d.GEMB_OK)
        raise ValidationError(msg)
    if notices or warnings:
        if submitting:
            msg = "Would you like to continue this submission?\n\n"
            dialog_type = c4d.GEMB_OKCANCEL
        else:
            msg = "Validate only.\n\n"
            dialog_type = c4d.GEMB_OK

        if warnings:
            msg += "Please check the warnings below:\n\n" + \
                "\n\n".join(["[WARN]:{}".format(w) for w in warnings]) + "\n\n"
        if notices:
            msg += "Please check the notices below:\n\n" + \
                "\n\n".join(["[INFO]:{}".format(n) for n in notices]) + "\n\n"

        result = c4d.gui.MessageDialog(msg, type=dialog_type)
        if result != c4d.GEMB_R_OK:
            c4d.WriteConsole("Submission cancelled by user.\n")
            raise ValidationError(msg)
    # Either there were no  messages, or the user clicked Continue (OK)


def _run_validators(dialog):
    # Eventually we will handle takes here
    takename = "Main"
    validators = [plugin(dialog) for plugin in Validator.plugins()]
    for validator in validators:
        validator.run(takename)

    errors = list(set.union(*[validator.errors for validator in validators]))
    warnings = list(
        set.union(*[validator.warnings for validator in validators]))
    notices = list(set.union(*[validator.notices for validator in validators]))
    return errors, warnings, notices




from ciocore.gpath_list import PathList
import importlib
import c4d

__data__ = None


def data(dialog):
    global __data__

    if __data__:
        return __data__
 
    __data__ = dialog.section("AssetsSection").get_assets_path_list()
 
    return __data__

def clear():
    global __data__
    __data__ = None

 



import c4d
from cioc4d.sections.collapsible_section import CollapsibleSection
from cioc4d.widgets.text_grp import TextGrp
from ciocore.sequence import Sequence

VALID_KEYS = [ "frame_spec" , "task_count", "frame_count", "scout_frame_spec", "scout_task_count", "scout_frame_count" ]
class InfoSection(CollapsibleSection):
    ORDER = 40

    def __init__(self, dialog):

        self.frame_info_widget = None
        self.scout_info_widget = None

        self.frame_spec=""
        self.task_count=0
        self.frame_count=0
  
        self.scout_frame_spec=""
        self.scout_task_count=0
        self.scout_frame_count=0


        super(InfoSection, self).__init__(dialog, "Info", collapse=False)

    def build(self):

        self.frame_info_widget = TextGrp(
            self.dialog, label="Frame info")

        self.scout_info_widget = TextGrp(
            self.dialog, label="Scout info")

    def on_plugin_message(self, widget_id, msg):
        """
        Update info if something changed in the Frames section.
        """
        frames_section = self.dialog.section("FramesSection")

        if widget_id in frames_section.get_preview_affectors():
            self.calculate_info()

    def on_core_message(self, msg_id, msg):
        """Repopulate if EVMSG_CHANGE was emitted.

        It's possibly a result of frame range edit or loading a new file.
        """
        if msg_id in [c4d.EVMSG_CHANGE]:
            self.calculate_info()

    def calculate_info(self):
        """Build the info strings using data that came from the frames section."""
        frames_section = self.dialog.section("FramesSection")
        try:
            sequence = frames_section.get_sequence()
        except (ValueError, TypeError):
            self.frame_info_widget.set_value("INVALID SEQUENCE")
            self.scout_info_widget.set_value("INVALID SEQUENCE")
            return
        frame_count = len(sequence)
        task_count = sequence.chunk_count()

        frames_info = "spec:{} --- tasks:{:d} --- frames:{:d}".format(
            sequence, task_count, frame_count)
        self.set_member_vars(frame_spec=str(sequence), task_count=task_count, frame_count=frame_count)
        scout_info = "No scout frames"



        scout_task_count = 0
        try:
            scout_sequence = frames_section.get_scout_sequence(sequence)
        except (ValueError, TypeError):
            scout_sequence = None
        scout_tasks_sequence = None
        if scout_sequence:
            scout_chunks = sequence.intersecting_chunks(scout_sequence)
            if scout_chunks:
                scout_tasks_sequence = Sequence.create(
                    ",".join(str(chunk) for chunk in scout_chunks))
                scout_task_count = len(scout_chunks)

                scout_frame_count =  len(scout_tasks_sequence)
                scout_info = "spec:{} --- tasks:{:d} --- frames:{:d}".format(
                    scout_tasks_sequence, scout_task_count, scout_frame_count)
                self.set_member_vars(scout_frame_spec=str(scout_tasks_sequence), scout_task_count=scout_task_count, scout_frame_count=scout_frame_count)
        self.frame_info_widget.set_value(frames_info)
        self.scout_info_widget.set_value(scout_info)
 
    def set_member_vars(self,**kwargs) :
        for key in kwargs:
            if key not in VALID_KEYS:
                continue
            setattr(self, key,kwargs[key]) 
        

       



import json
import os
import c4d
import cioc4d.const as k
from cioc4d.conductor_store import ConductorStore
from cioc4d.foot_grp import FootGroup
from cioc4d.head_grp import HeadGroup
# NOTE: SECTIONS MUST BE IMPORTED, EVEN THOUGH THEY ARE NOT CALLED BY NAME
from cioc4d.sections.assets_section import AssetsSection
from cioc4d.sections.autosave_section import AutosaveSection
from cioc4d.sections.collapsible_section import CollapsibleSection
from cioc4d.sections.diagnostics_section import DiagnosticsSection
from cioc4d.sections.environment_section import EnvironmentSection
from cioc4d.sections.frames_section import FramesSection
from cioc4d.sections.general_section import GeneralSection
from cioc4d.sections.info_section import InfoSection
from cioc4d.sections.location_section import LocationSection
from cioc4d.sections.metadata_section import MetadataSection
from cioc4d.sections.notification_section import NotificationSection
from cioc4d.sections.preview_section import PreviewSection
from cioc4d.sections.retries_section import RetriesSection
from cioc4d.sections.software_section import SoftwareSection
from cioc4d.sections.task_section import TaskSection
from cioc4d.sections.upload_options_section import UploadOptionsSection
# End of Sections
from ciocore import data as coredata
from ciocore.expander import Expander

from cioc4d import submit
from ciocore.validator import ValidationError
from cioc4d import validation
from ciocore.gpath import Path

class ConductorDialog(c4d.gui.GeDialog):
    """
    A dialog composed of sections of UI.

    In between a header and a footer, the dialog displays all sections inherited
    from CollapsibleSection according to their ORDER property. See
    cioc4d.sections.collapsible_section for inheritance requirements.

    Sections are designed to keep their logic in one place. 
    """

    def __init__(self):
        """Initialize the dialog with a list of section classes to be instantiated."""
        self.AddGadget(c4d.DIALOG_NOMENUBAR, 0)
        self.head_grp = None
        self.foot_grp = None
        self.next_id = 20000

        self._section_classes = sorted(
            CollapsibleSection.__subclasses__(), key=lambda x: x.ORDER)

        self.sections = []
        self.store = None

        coredata.init(product="cinema4d")

    # c4d method

    def CreateLayout(self):
        """
        Called by c4d to build the layout.

        We instantiate all collapsible sections, and put all except the preview
        section in a scroll layout. They are passed a reference to this dialog
        so they may register widgets and so on.
        """
        self.sections = []
        self.SetTitle("Conductor Render")
        self.head_grp = HeadGroup(self)
        self.AddSeparatorH(inith=0)
        scroll_grp = self.ScrollGroupBegin(
            id=self.register(),
            flags=c4d.BFH_SCALEFIT | c4d.BFV_SCALEFIT,

            scrollflags=c4d.SCROLLGROUP_VERT)
        if scroll_grp:
            main_grp = self.GroupBegin(
                id=self.register(),
                flags=c4d.BFH_SCALEFIT | c4d.BFV_SCALEFIT,
                title="",
                rows=1,
                cols=1,
                groupflags=0)

            if main_grp:
                self.GroupBorderSpace(4, 4, 4, 4)
                self.sections = [
                    cls(self) for cls in self._section_classes if cls.__name__ is not "PreviewSection"
                ]
            self.GroupEnd()  # manin
        self.GroupEnd()  # scroll

        self.AddSeparatorH(inith=3, flags=c4d.BFH_FIT | c4d.BFV_TOP)

        self.sections.append(PreviewSection(self))

        self.AddSeparatorH(inith=0)

        self.foot_grp = FootGroup(self)

        return True

    # c4d method
    def InitValues(self):
        """Call each section's populate_from_store method."""
        self.store = ConductorStore()
        self.populate_all_sections()

        return True

    # c4d method
    def Command(self, widget_id, msg):
        """

        Receive all events from this plugin.

        Tell the preview section the list of affectors that can influence it.
        This must be the first step in this Command callback because some
        on_plugin_message() functions may delete widgets, meaning those
        widget_ids won't be available to query later by
        PreviewSection::on_plugin_message().

              2.It could be optimized to only care about dynamic widgets!!
        if widget_id == self.foot_grp.save_button_id:
            self.save_to_store()
            return
        """
        if widget_id == self.foot_grp.submit_button_id:
            self.on_submit()
            return True
        if widget_id == self.foot_grp.validate_button_id:
            self.on_validate()
            return True

        for section in self.sections:
            section.set_store_affectors()

        self.section("PreviewSection").set_affectors()

        if widget_id == self.foot_grp.connect_button_id:
            self.on_connect()

        # Let all sections know about the event
        for section in self.sections:
            section.on_plugin_message(widget_id, msg)

        return True

    # c4d method
    def CoreMessage(self, msg_id, msg):
        """
        All the sections receive all EVMSG_CHANGE events from c4d and they decide what to do with them.
        """

        if msg_id == c4d.EVMSG_CHANGE:

            document_was_changed = self.store.on_scene_change()

            if document_was_changed:

                self.populate_all_sections()

            for section in self.sections:
                section.on_core_message(msg_id, msg)

        return super(ConductorDialog, self).CoreMessage(msg_id, msg)

    # c4d method
    def Message(self, msg, result):
        """
        All the sections receive mouse events from c4d and they decide what to do with them.
        """
        should_handle_ids = [c4d.BFM_INPUT]
        msg_id = msg.GetId()
        if msg_id in should_handle_ids:
            for section in self.sections:
                section.on_message(msg_id, msg)

        return super(ConductorDialog, self).Message(msg, result)

    # c4d method
    def AskClose(self):
        return False

    def section(self, classname):
        """
        Convenience to allow sections to find other sections by name.

        See how the InfoSection finds the FramesSection so it can call it's methods.
        """
        return next(s for s in self.sections if s.__class__.__name__ == classname)

    def register(self):
        """
        Register a UI Element and keep track of the next available ID.

        This, in conjunction with the object oriented structure, avoids the
        tedious job of tracking explicit IDs.
        """
        current_id = self.next_id
        self.next_id += 1
        return current_id

    def on_connect(self):

        try:
            try_cache = self.section(
                "DiagnosticsSection").fixture_widget.get_value()
            coredata.data(force=True, try_cache=try_cache)
            self.foot_grp.enable_submit(True)
            return
        except BaseException:
            c4d.WriteConsole(
                "Try again after deleting your credentials file (~/.config/conductor/credentials)")
            raise
        self.foot_grp.enable_submit(False)

    def calculate_submission(self, **kwargs):
        """
        Ask each section to contribute to the submission object.
        """

        context = self.get_context()
        expander = Expander(safe=True, **context)
        submission = {}

        for section in self.sections:
            submission.update(section.resolve(expander, **kwargs))

        return submission

    def get_context(self):
        doc = c4d.documents.GetActiveDocument()
        docname = doc.GetDocumentName()
        docnamex, ext = os.path.splitext(doc.GetDocumentName())
        docdir = doc.GetDocumentPath()
        if docdir:
            docdir = Path(docdir).posix_path(with_drive=False)
        docfile = os.path.join(docdir, docname)
        try:
            takename = doc.GetTakeData().GetCurrentTake().GetName()
        except AttributeError:
            takename = "Main"
        result = {
            "docname": docname,
            "docnamex": docnamex,
            "docext": ext,
            "docdir": docdir,
            "docfile": docfile,
            "takename": takename
        }
        return result

    def populate_all_sections(self):
        for section in self.sections:
            section.populate_from_store()
        self.section("PreviewSection").set_submission_preview()

    def on_submit(self):
        submit.submit(self)

    def on_validate(self):
        try:
            validation.run(self, submitting=False)
        except ValidationError as ex:
            c4d.WriteConsole(str(ex))
